# encoding: utf-8
"""
Tools to work with DVM paramater list.
"""

from hit_acs.util import csv_unicode_reader


# CSV column types

def CsvStr(s):
    return s


def CsvInt(s):
    return int(s) if s else None


def CsvFloat(s):
    return float(s) if s else None


def CsvUnit(s):
    s = s.replace(u'grad', u'degree')
    s = s.replace(u'Ohm', u'ohm')
    s = s.replace(u'part.', u'count')   # used for particle count
    return s


def load_csv(lines, encoding='utf-8', delimiter=';'):
    """
    Parse DVM parameters from CSV file exported from XLS documentation
    spreadsheet (e.g. DVM-Parameter_v2.10.0-10-HIT.xls)
    """
    return load_csv_data(csv_unicode_reader(
        lines, encoding=encoding, delimiter=delimiter))


def load_csv_data(rows):
    def parse_row(row):
        return {
            n: _csv_column_types[n](row[i].strip())
            for n, i in _csv_column_index.items()
        }
    return map(parse_row, rows)


# all columns in csv file:
_csv_column_names = [
    '',                 # Nr. für Link
    'name',             # Code Param (GSI-Nomenklatur)
    '',                 # Code Gerät (GSI- NomenkLatur) entspr. DCU!
    '',                 # Code Gruppe (=Kalkulationsgruppe); möglichst
                        #       GSI-NomenkLatur
    'ui_name',          # GUI Beschriftung Parameter (ohne Einheit)
    'ui_hint',          # GUI Beschriftung Hint
    '',                 # Position ExpertGrids
    '',                 # DVM liest Parameter
    '',                 # DVM ändert Parameter
    '',                 # DVM Datensatz spezifisch
    '',                 # Rein temporär
    '',                 # MEFI-Abhängigkeit
    '',                 # Input Param wird Output Param bei MEFI
    '',                 # In Gui Init änderbar
    '',                 # Daten-typ
    'ui_prec',          # Präzision (Anz. Nachkomma im GUI)
    'unit',             # Einheit Parameter
    'ui_unit',          # Einheit Anzeige im GUI
    'ui_conv',          # Umrechnungsfaktor Einheit--> Einheit GUI
    '',                 # Beispielwert für Test in Einheit GUI
    '',                 # Referenz auf DCU /MDE
    '',                 # (nicht verwendet)
    '',                 # Zugriffscode / editierbarkeit
    '',                 # Versions-  Relevanz
    '',                 # Detail Ansicht verfügbar (ja/nein)
    '',                 # Link auf Maximalwert
    '',                 # Link auf Minimalwert
    '',                 # Code Min/Max- Rechen-vorschrift
    '',                 # Master-gruppe
    '',                 # Defaultwert Änderung pro Pfeiltasten-druck/
                        #       Maus-radsegment in Einheit GUI
    '',                 # Im laufenden Betrieb änderbar (ja/ nein)
    '',                 # Link auf zugehörigen sekundären Wert
]

_csv_column_types = {
    'name':     CsvStr,
    'ui_name':  CsvStr,
    'ui_hint':  CsvStr,
    'ui_prec':  CsvInt,
    'unit':     CsvUnit,
    'ui_unit':  CsvUnit,
    'ui_conv':  CsvFloat,
}

# inverse map of _csv_columns[i] (used columns)
_csv_column_index = {
    name: index
    for index, name in enumerate(_csv_column_names)
    if name
}



# encoding: utf-8
"""
Madgui online control plugin.
"""

from __future__ import absolute_import

import os
import logging

from importlib_resources import read_binary
from pydicti import dicti

from .beamoptikdll import BeamOptikDLL, ExecOptions
from .beamoptikstub import BeamOptikStub

import madgui.util.unit as unit
import madgui.online.api as api
from madgui.util.misc import relpath as safe_relpath
from madgui.util.collections import Bool
from madgui.util.qt import SingleWindow

from .dvm_parameters import load_csv
from .offsets import find_offsets


ENERGY_PARAM = {
    'lebt': 'E_SOURCE',
    'mebt': 'E_MEBT',
}

PERIODIC_TABLE = {
    1: 'p',
    2: 'He',
    6: 'C',
    8: 'O',
}

MEFI_PARAMS = ('beam_energy', 'beam_focus', 'beam_intensity', 'gantry_angle')


def load_dvm_parameters():
    blob = read_binary('hit_acs', 'DVM-Parameter_v2.10.0-HIT.csv')
    parlist = load_csv(blob.splitlines(), 'utf-8')
    return dicti({p['name']: p for p in parlist})


class _HitACS(api.Backend):

    def __init__(self, lib, params, model=None, offsets=None, settings=None,
                 control=None):
        self._lib = lib
        self._params = dicti({
            'beam_energy': dict(
                name='beam_energy',
                ui_name='beam_energy',
                ui_hint='',
                ui_prec=3,
                unit='MeV/u',
                ui_unit='MeV/u',
                ui_conv=1),
            'beam_focus': dict(
                name='beam_focus',
                ui_name='beam_focus',
                ui_hint='',
                ui_prec=3,
                unit='m',
                ui_unit='mm',
                ui_conv=1000),
            'beam_intensity': dict(
                name='beam_intensity',
                ui_name='beam_intensity',
                ui_hint='',
                ui_prec=3,
                unit='',
                ui_unit='',
                ui_conv=1),
            'gantry_angle': dict(
                name='gantry_angle',
                ui_name='gantry_angle',
                ui_hint='',
                ui_prec=3,
                unit='°',
                ui_unit='°',
                ui_conv=1),
        })
        self._params.update(params)
        self._model = model
        self._offsets = {} if offsets is None else offsets
        self.connected = Bool(False)
        self.settings = settings
        self.control = control
        self.vAcc = 0

    @property
    def beamoptikdll(self):
        """Python wrapper for the BeamOptikDLL exposed methods."""
        return self._lib

    # Backend API

    def connect(self):
        """Connect to online database (must be loaded)."""
        self._lib.GetInterfaceInstance()
        self.connected.set(True)
        self.vAcc = self._lib.GetSelectedVAcc()

    def disconnect(self):
        """Disconnect from online database."""
        (self.settings or {}).update(self.export_settings())
        self._lib.FreeInterfaceInstance()
        self.connected.set(False)

    def export_settings(self):
        """Updates the settings yaml file for future loggins"""
        mefi = self._lib.GetMEFIValue()[1]
        settings = {
            'variant': self._lib._variant,
            'vacc': self._lib.GetSelectedVAcc(),
            'mefi': mefi and tuple(mefi),
        }
        if hasattr(self._lib, 'export_settings'):
            settings.update(self._lib.export_settings())
        return settings

    def execute(self, options=ExecOptions.CalcDif):
        """Execute changes (commits prior set_value operations)."""
        self._lib.ExecuteChanges(options)
        self.update_model_with_vAcc()

    def param_info(self, knob):
        """Get parameter info for backend key."""
        data = self._params.get(knob)
        return data and api.ParamInfo(**data)

    def read_monitor(self, name):
        """
        Read out one monitor, return values as dict with keys
        posx/posy/envx/envy.
        """
        # TODO: Handle usability of parameters individually
        try:
            GetFloatValueSD = self._lib.GetFloatValueSD
            posx = GetFloatValueSD('posx_' + name)
            posy = GetFloatValueSD('posy_' + name)
            envx = GetFloatValueSD('widthx_' + name)
            envy = GetFloatValueSD('widthy_' + name)
        except RuntimeError:
            return {}
        # TODO: move sanity check to later, so values will simply be
        # unchecked/grayed out, instead of removed completely
        # The magic number -9999.0 signals corrupt values.
        # FIXME: Sometimes width=0 is returned. ~ Meaning?
        if posx == -9999 or posy == -9999 or envx <= 0 or envy <= 0:
            return {}
        xoffs, yoffs = self._offsets.get(name, (0, 0))
        return {
            'posx': -(posx / 1000 + xoffs),
            'posy': +(posy / 1000 + yoffs),
            'envx': envx / 1000,
            'envy': envy / 1000,
        }

    def read_params(self, param_names=None, warn=True):
        """Read all specified params (by default all). Return dict."""
        if param_names is None:
            param_names = self._params
            warn = False
        return {
            param: value
            for param in param_names
            for value in [self.read_param(param, warn=warn)]
            if value is not None
        }

    def read_param(self, param, warn=True):
        """Read parameter. Return numeric value."""
        param = param.lower()
        if param in MEFI_PARAMS:
            return self._lib.GetMEFIValue()[0][MEFI_PARAMS.index(param)]
        try:
            return self._lib.GetFloatValue(param)
        except RuntimeError as e:
            if warn:
                logging.warning("{} for {!r}".format(e, param))

    def write_param(self, param, value):
        """Update parameter into control system."""
        param = param.lower()
        if param in MEFI_PARAMS:
            cur_value = self.read_param(param)
            if value != cur_value:
                logging.warning(
                    "Unable to set {}={} (is {}). This parameter "
                    "can only be changed by selecting the MEFI combination!"
                    .format(param, value, cur_value))
            return
        try:
            self._lib.SetFloatValue(param, value)
        except RuntimeError as e:
            logging.error("{} for {!r} = {}".format(e, param, value))

    def get_beam(self):
        units  = unit.units
        e_para = ENERGY_PARAM.get(self._model().seq_name, 'E_HEBT')
        z_num  = self._lib.GetFloatValue('Z_POSTSTRIP')
        mass   = self._lib.GetFloatValue('A_POSTSTRIP') * units.u
        charge = self._lib.GetFloatValue('Q_POSTSTRIP') * units.e
        e_kin  = (self._lib.GetFloatValue(e_para) or 1) * units.MeV / units.u
        return {
            'particle': PERIODIC_TABLE[round(z_num)],
            'charge':   unit.from_ui('charge', charge),
            'mass':     unit.from_ui('mass',   mass),
            'energy':   unit.from_ui('energy', mass * (e_kin + 1*units.c**2)),
        }

    def vAcc_to_model(self):
        """User defined vAcc to model"""
        # No work around this unless
        # we implement a yml config file with
        # this information (as Thomas would have done it)
        # TODO: Implement a yaml file to keep code style
        T1 = [1, 6,  11]
        T2 = [2, 7,  12]
        GA = [3, 8,  13]
        QS = [4, 9,  14]
        BD = [5, 10, 15]
        if (self.vAcc in T1):
            return 'hht1.cpymad.yml'
        if (self.vAcc in T2):
            return 'hht2.cpymad.yml'
        if (self.vAcc in GA):
            return 'hht3.cpymad.yml'
        if (self.vAcc in QS):
            return 'hht4.cpymad.yml'
        if (self.vAcc in BD):
            return 'hht5.cpymad.yml'
        logging.warning('vAcc is not standard. Load model manually.')
        return self.model().model_data()['sequence']

    def update_model_with_vAcc(self):
        """If vAcc was changed, ask if the model also should be changed"""
        new_vAcc = self._lib.GetSelectedVAcc()
        same_vAcc = (self.vAcc == new_vAcc)
        if (not same_vAcc):
            self.vAcc = new_vAcc
            self.control.auto_load_model()


class HitACS(_HitACS):

    def __init__(self, session, settings):
        """Connect to online database."""
        params = load_dvm_parameters()
        offsets = find_offsets(settings.get('runtime_path', '.'))
        lib = session.user_ns.beamoptikdll = BeamOptikDLL(
            variant=settings.get('variant', 'HIT'))
        super().__init__(lib, params, session.model, offsets, settings,
                         session.control)


class TestACS(_HitACS):

    def __init__(self, session, settings):
        params = load_dvm_parameters()
        offsets = find_offsets(settings.get('runtime_path', '.'))
        # Don't pass `session.model()` to the stub. It should use an
        # independent simulation, which is cloned upon connection in
        # `on_model_changed`:
        lib = session.user_ns.beamoptikdll = BeamOptikStub(
            None, offsets, settings)
        super().__init__(lib, params, session.model, offsets,
                         control=session.control)
        self.menu = None
        self.window = None
        self.set_window(session.window())
        self.connected.changed.connect(self.on_connected_changed)

        self.str_file = settings.get('str_file')
        self.sd_file = settings.get('sd_file')

    def load_float_values(self, filename):
        from madgui.util.export import read_str_file
        self.str_file = filename = os.path.abspath(filename)
        self._lib.set_float_values(read_str_file(filename))

    def load_sd_values(self, filename):
        from madgui.util.yaml import load_file
        self.sd_file = filename = os.path.abspath(filename)
        data = load_file(filename)
        cols = {
            'envx': 'widthx',
            'envy': 'widthy',
            'x': 'posx',
            'y': 'posy',
        }
        self.auto_sd.set(False)
        self._lib.set_sd_values({
            cols[param]+'_'+elem: value
            for elem, values in data['monitor'].items()
            for param, value in values.items()
        })

    def set_window(self, window):
        self.window = window
        self.menu = window and window.acs_settings_menu
        if window is None:
            return
        from madgui.util.menu import extend, Item, Separator
        self.jitter = Bool(self._lib.jitter)
        self.auto_sd = Bool(self._lib.auto_sd)
        self.menu.clear()
        extend(window, self.menu, [
            Item('&Vary readouts', None,
                 'Emulate continuous readouts using gaussian jitter',
                 self._toggle_jitter,
                 checked=self.jitter),
            Item('Add &magnet aberrations', None,
                 'Add small deltas to all magnet strengths',
                 self._lib._aberrate_strengths),
            Separator,
            Item('Autoset readouts from model', None,
                 'Autoset monitor readout values from model twiss table',
                 self._toggle_auto_sd,
                 checked=self.auto_sd),
            Separator,
            Item('Load readouts from file', None,
                 'Load monitor readout values from monitor export',
                 self._open_sd_values),
            Item('Load strengths from file', None,
                 'Load magnet strengths from strength export',
                 self._open_float_values),
            Separator,
            Item('Edit backend simulation model', None,
                 'Edit initial conditions for the model used to simulate '
                 'the backend behaviour',
                 self._edit_model_initial_conditions.create),
        ])

    def export_settings(self):
        return {
            'jitter': self.jitter(),
            'shot_interval': self._lib.sd_cache.timeout,
            'auto_sd': self.auto_sd(),
            'str_file': self.str_file and safe_relpath(self.str_file, None),
            'sd_file': self.str_file and safe_relpath(self.sd_file, None),
        }

    def _toggle_jitter(self):
        self.jitter.set(not self.jitter())
        self._lib.jitter = self.jitter()

    def _toggle_auto_sd(self):
        self.auto_sd.set(not self.auto_sd())
        self._lib.auto_sd = self.auto_sd()
        self._lib.update_sd_values()

    def _open_sd_values(self):
        from madgui.widget.filedialog import getOpenFileName
        filters = [
            ("YAML files", "*.yml", "*.yaml"),
            ("All files", "*"),
        ]
        folder = self.window.str_folder or self.window.folder
        if self.sd_file:
            folder = os.path.dirname(self.sd_file)
        filename = getOpenFileName(
            self.window, 'Open monitor export', folder, filters)
        if filename:
            self.load_sd_values(filename)

    def _open_float_values(self):
        from madgui.widget.filedialog import getOpenFileName
        filters = [
            ("STR files", "*.str"),
            ("All files", "*"),
        ]
        folder = self.window.str_folder or self.window.folder
        if self.str_file:
            folder = os.path.dirname(self.str_file)
        filename = getOpenFileName(
            self.window, 'Open strength export', folder, filters)
        if filename:
            self.load_float_values(filename)

    def on_connected_changed(self, connected):
        if connected:
            self.model.changed.connect(self.on_model_changed)
            self.on_model_changed(self.model())
        else:
            self.model.changed.disconnect(self.on_model_changed)
        if self.menu:
            self.menu.setEnabled(connected)

    def on_model_changed(self, model):
        clone = model and model.load_file(model.filename, stdout=False)
        self._lib.set_model(clone)
        if clone:
            if self.str_file:
                self.load_float_values(self.str_file)
            if self.sd_file:
                self.load_sd_values(self.sd_file)

    @SingleWindow.factory
    def _edit_model_initial_conditions(self):
        from madgui.widget.params import model_params_dialog
        return model_params_dialog(
            self._lib.model, parent=self.window, folder=self.window.folder)

    @property
    def model(self):
        return self._model



"""
Utilities for reading .xml file with MWPC offsets.
"""

import os
from glob import glob
from xml.etree import ElementTree

from madgui.util.unit import from_ui, from_config


PREFIX_ROOM = {'Room1': 'T1', 'Room2': 'T2', 'Room3': 'T3', 'Room4': 'T4'}
SUFFIX_MWPC = {'MWPC 1': 'DG1G', 'MWPC 2': 'DG2G', 'MWPC 3': 'DF1'}


def parse_datum(name, datum):
    value = float(datum.text)
    unit = from_config(datum.attrib['Unit'])
    return from_ui(name, unit, value)


def read_offsets_file(path):
    tree = ElementTree.parse(path)
    root = tree.getroot()

    abstract, calibration = root.iter('Table')
    metadata = {datum.attrib['ID']: datum.text
                for datum in abstract.iter('Data')}

    # Columns are:
    # - 0 = device name
    # - 1 = old X offset
    # - 2 = new X offset
    # - 3 = old Y offset
    # - 4 = new Y offset
    offsets = {
        row[0].text: (
            parse_datum('x', row[2][0]),
            parse_datum('y', row[4][0]))
        for row in calibration[1:]
    }

    room = metadata['TreatmentRoom']
    prefix = PREFIX_ROOM[room]
    return {prefix+SUFFIX_MWPC[mwpc]: xy
            for mwpc, xy in offsets.items()}


def find_offsets(path):
    """Find and read .xml files MWPC offsets in `path`."""
    offsets = {}
    for path in glob(os.path.join(path, '*', '*.xml')):
        try:
            offsets.update(read_offsets_file(path))
        except Exception:
            pass
    return offsets


def print_offsets(path):
    print(path+':')
    for mwpc, (x, y) in read_offsets_file(path).items():
        print("  {:6}: {:+.3f} mm, {:+.3f} mm".format(mwpc, x*1000, y*1000))


if __name__ == '__main__':
    import sys
    for path in sys.argv[1:]:
        print_offsets(path)



# encoding: utf-8
"""
Python wrapper for 'BeamOptikDLL.dll'.
"""

# NOTE: This module deliberately depends only on the standard library and can
# therefore be simply included in another application without having to
# install or provide anything else (except for the actual DLL of course).

from collections import namedtuple
from ctypes import c_double as Double, c_int as Int, POINTER
import ctypes
import logging
import platform

is_64bit = platform.architecture()[0] == '64bit'

try:
    basestring
except NameError:
    basestring = str


def _encode(s):
    return s if isinstance(s, bytes) else s.encode('utf-8')


def _decode(s):
    return s.decode('utf-8') if isinstance(s, bytes) else s


if is_64bit:
    def Str(s):
        return _Str(_decode(s))
    _Str = ctypes.c_wchar_p         # constructor wants unicode
else:
    def Str(s):
        return _Str(_encode(s))
    _Str = ctypes.c_char_p          # constructor wants bytes


try:
    NewValueCallback = ctypes.WINFUNCTYPE(
        None, ctypes.c_char_p, POINTER(Double), POINTER(Int))
except AttributeError:
    NewValueCallback = None


EFI = namedtuple('EFI', ['energy', 'focus', 'intensity', 'gantry_angle'])


class _EnumBase(int):

    """Abstract base type for enums (missed :cvar:`_value_names`)."""

    def __repr__(self):
        return '<{}.{}: {}>'.format(
            self.__class__.__name__,
            self._value_names[int(self)],
            int(self),
        )


def make_enum(name, value_names):
    """Create a simple enum type (like in C++)."""
    Enum = type(name, (_EnumBase,), {
        '_value_names': value_names,
    })
    for i, v in enumerate(value_names):
        setattr(Enum, v, Enum(i))
    return Enum


DVMStatus = make_enum('DVMStatus', ['Stop', 'Idle', 'Init', 'Ready',
                                    'Busy', 'Finish', 'Error'])
GetOptions = make_enum('GetOptions', ['Current', 'Saved'])
ExecOptions = make_enum('ExecOptions', ['CalcAll', 'CalcDif', 'SimplyStore'])
GetSDOptions = make_enum('GetSDOptions', ['Current', 'Database', 'Test'])


class BeamOptikDLL(object):

    """
    Thin wrapper around the BeamOptikDLL API.

    It abstracts the ctypes data types and automates InterfaceId as well as
    iDone. Nothing else.

    You must call the ``GetInterfaceInstance`` method after object creation to
    initialize the DLL.

    >>> obj = BeamOptikDLL()
    >>> obj.GetInterfaceInstance()
    """

    filename = 'BeamOptikDLL64.dll' if is_64bit else 'BeamOptikDLL.dll'

    def __init__(self, lib=filename, variant='HIT'):
        """
        Load library and initialize member variables.

        :param str lib: filename or DLL proxy object
        :param str variant: 'HIT' or 'MIT', decides whether the `_RKA` set of
                            functions will be used internally
        """
        if isinstance(lib, basestring):
            lib = ctypes.windll.LoadLibrary(lib)
        self.lib = lib
        self._funcs = _load_functions(lib)
        self._iid = None
        self._variant = variant

    def __bool__(self):
        """Check if the object belongs to an initialized interface instance."""
        return self._iid is not None

    __nonzero__ = __bool__

    @property
    def iid(self):
        """Interface instance ID."""
        if self._iid is None:
            raise RuntimeError(
                "GetInterfaceInstance must be called "
                "before using other methods.")
        return self._iid

    def DisableMessageBoxes(self):
        """
        Prevent creation of certain message boxes.

        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('DisableMessageBoxes')

    def GetInterfaceInstance(self):
        """
        Initialize a BeamOptikDLL instance (connects DB and initialize DLL).

        :return: new instance id
        :rtype: int
        :raises RuntimeError: if the exit code indicates any error
        """
        if self._iid is not None:
            raise RuntimeError("GetInterfaceInstance cannot be called twice.")
        iid = Int()
        self._call('GetInterfaceInstance', iid)
        self._iid = iid
        return iid.value

    def FreeInterfaceInstance(self):
        """
        Free resources.

        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('FreeInterfaceInstance', self.iid)
        self._iid = None

    def GetDVMStatus(self):
        """
        Get current status of selected virtual accelerator.

        :return: DVM status
        :rtype: DVMStatus
        :raises RuntimeError: if the exit code indicates any error
        """
        status = Int()
        self._call('GetDVMStatus', self.iid, status)
        return DVMStatus(status.value)

    def SelectVAcc(self, vaccnum):
        """
        Select the virtual accelerator.

        :param int vaccnum: virtual accelerator number (0-255)
        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('SelectVAcc', self.iid, Int(vaccnum))

    def SelectMEFI(self, vaccnum, energy, focus, intensity, gantry_angle=0):
        """
        Select EFI combination for the currently selected VAcc.

        :param int vaccnum: virtual accelerator number (0-255)
        :param int energy: energy channel (1-255)
        :param int focus: focus channel (1-6)
        :param int intensity: intensity channel (1-15)
        :param int gantry_angle: HIT: gantry angle index (1-36)
                                 MIT: extraction time
        :return: physical EFI values
        :rtype: EFI
        :raises RuntimeError: if the exit code indicates any error

        CAUTION: SelectVAcc must be called before invoking this function!
        """
        values = [Double(), Double(), Double(), Double()]
        func = ('SelectMEFI', 'SelectMEFI_RKA')[self._variant == 'MIT']
        self._call(func, self.iid, Int(vaccnum),
                   Int(energy), Int(focus), Int(intensity), Int(gantry_angle),
                   *values)
        return EFI(*[v.value for v in values])

    def GetSelectedVAcc(self):
        """
        Get selected virtual accelerator.

        :return: virtual accelerator number (0-255)
        :rtype: int
        :raises RuntimeError: if the exit code indicates any error
        """
        vaccnum = Int()
        self._call('GetSelectedVAcc', self.iid, vaccnum)
        return vaccnum.value

    def GetFloatValue(self, name, options=GetOptions.Current):
        """
        Get parameter value.

        :param str name: parameter name
        :param GetOptions options: options
        :return: parameter value
        :rtype: float
        :raises RuntimeError: if the exit code indicates any error
        """
        value = Double()
        self._call('GetFloatValue', self.iid, Str(name), value, Int(options))
        return value.value

    def SetFloatValue(self, name, value, options=0):
        """
        Set parameter value.

        :param str name: parameter name
        :param float value: parameter value
        :param options: not used currently
        :raises RuntimeError: if the exit code indicates any error

        Changes take effect after calling :func:`ExecuteChanges`.
        """
        self._call('SetFloatValue', self.iid,
                   Str(name), Double(value), Int(options))

    def ExecuteChanges(self, options=ExecOptions.CalcDif):
        """
        Apply parameter changes.

        :param ExecOptions options: what to do exactly
        :raises RuntimeError: if the exit code indicates any error
        """
        self._call('ExecuteChanges', self.iid, Int(options))

    def SetNewValueCallback(self, callback):
        """
        Install a callback function that should be called whenever the
        calculation of a new value is complete (…whatever that means).

        :param callback: ``callable(name:str, val:float, type:int)``
        :raises RuntimeError: if the exit code indicates any error
        """
        def c_callback(name, value, type_):
            return callback(_decode(name),
                            value.contents.value,
                            type_.contents.value)
        # store a reference to keep the callback object alive:
        self._c_cb = NewValueCallback(0 if callback is None else c_callback)
        self._call('SetNewValueCallback', self.iid, self._c_cb)

    def GetFloatValueSD(self, name, options=GetSDOptions.Current):
        """
        Get current beam measurement at specific element.

        :param str name: parameter name (<observable>_<element name>)
        :param GetSDOptions options: options
        :return: measured value
        :rtype: float
        :raises RuntimeError: if the exit code indicates any error
        """
        value = Double()
        self._call('GetFloatValueSD', self.iid, Str(name), value, Int(options))
        return value.value

    def GetLastFloatValueSD(self, name, vaccnum,
                            energy, focus, intensity, gantry_angle=0,
                            options=GetSDOptions.Current):
        """
        Get previous beam measurement at specific element.

        :param str name: parameter name (<observable>_<element name>)
        :param int vaccnum: virtual accelerator number (0-255)
        :param GetSDOptions options: options
        :return: measured value and EFI combination
        :rtype: tuple
        :raises RuntimeError: if the exit code indicates any error
        """
        value = Double()
        func = ('GetLastFloatValueSD' if self.variant == 'HIT' else
                'GetLastFloatValueSD_RKA')
        self._call(func, self.iid, Str(name),
                   value, Int(vaccnum), Int(options),
                   Int(energy), Int(focus), Int(intensity), Int(gantry_angle))
        return value.value

    def StartRampDataGeneration(self, vaccnum, energy, focus, intensity):
        """
        Call StartRampDataGeneration().
        """
        order_num = Int()
        self._call('StartRampDataGeneration', self.iid,
                   Int(vaccnum), Int(energy), Int(focus), Int(intensity),
                   order_num)
        return order_num.value

    def GetRampDataValue(self, order_num, event_num, delay,
                         parameter_name, device_name):
        """
        Call GetRampDataValue()
        """
        value = Double()
        self._call('GetRampDataValue', self.iid,
                   Int(order_num), Int(event_num), Int(delay),
                   Str(parameter_name), Str(device_name),
                   value)
        return value.value

    def SetIPC_DVM_ID(self, name):
        """Call SetIPC_DVM_ID(). Not implemented!"""
        raise NotImplementedError()     # TODO

    def GetMEFIValue(self):
        """
        Retrieve EFI values for current selection.

        :return: physical EFI values, EFI channel numbers
        :rtype: tuple(EFI, EFI)
        :raises RuntimeError: if the exit code indicates any error
        """
        if self._variant == 'HIT':
            values = [Double(), Double(), Double(), Double()]
            channels = [Int(), Int(), Int(), Int()]
            self._call('GetMEFIValue', self.iid, *(values + channels))
            return (EFI(*[v.value for v in values]),
                    EFI(*[c.value for c in channels]))
        elif self._variant == 'MIT':
            values = [Double(), Double(), Double(), Double()]
            self._call('GetMEFIValue_RKA', self.iid, *values)
            return (EFI(*[v.value for v in values]), None)

    # internal methods

    def _call(self, function, *params):
        """
        Call the specified DLL function.

        :param str function: name of the function to call
        :param params: ctype function parameters except for piDone.
        :raises RuntimeError: if the exit code indicates any error

        For internal use only!
        """
        done = Int()
        params = list(params)
        if function in ('SelectMEFI', 'SelectMEFI_RKA'):
            params.insert(6, done)
        else:
            params.append(done)
        if function != 'GetFloatValueSD':
            logging.debug('{}{}'.format(function, tuple(params)))
        func = self._funcs[function]
        func(*params)
        self.check_return(done.value)

    @classmethod
    def check_return(cls, done):
        """
        Check DLL-API exit code for errors and raise exception.

        :param int done: exit code of an DLL function
        :raises RuntimeError: if the exit code is a known error code != 0
        :raises ValueError: if the exit code is unknown
        """
        if 0 < done and done < len(cls.error_messages):
            raise RuntimeError(cls.error_messages[done])
        elif done != 0:
            raise ValueError("Unknown error: %i" % done)

    error_messages = [
        None,
        "Invalid Interface ID.",
        "Parameter not found in internal DVM list.",
        "GetValue failed.",
        "SetValue failed.",
        "Unknown option.",
        "Memory error.",
        "General runtime error.",
        "Ramp event not supported.",
        "Ramp data not available.",
        "Invalid offset for ramp function."]


def _load_functions(lib):
    """Load the function pointers for all exported functions and
    initialize their argtypes. Return as dict ``{name: function}``."""
    i = POINTER(Int)
    d = POINTER(Double)
    s = _Str
    return _declare(lib, {
        'GetInterfaceInstance':     [i, i],
        'FreeInterfaceInstance':    [i, i],
        'DisableMessageBoxes':      [i],
        'GetDVMStatus':             [i, i, i],
        'SelectVAcc':               [i, i, i],
        'SelectMEFI':               [i, i, i, i, i, i, i, d, d, d, d],
        'SelectMEFI_RKA':           [i, i, i, i, i, i, i, d, d, d, d],
        'SelectMEFI_EXT':           [i, i, i, i, i, i, i, i, d, d, d, d],
        'SelectMEFI_EXT_RKA':       [i, i, i, i, i, i, i, i, d, d, d, d],
        'GetSelectedVAcc':          [i, i, i],
        'GetFloatValue':            [i, s, d, i, i],
        'SetFloatValue':            [i, s, d, i, i],
        'ExecuteChanges':           [i, i, i],
        'SetNewValueCallback':      [i, NewValueCallback, i],
        'GetFloatValueSD':          [i, s, d, i, i],
        'GetLastFloatValueSD':      [i, s, d, i, i, i, i, i, i, i],
        'GetLastFloatValueSD_RKA':  [i, s, d, i, i, i, i, i, i, i],
        'StartRampDataGeneration':  [i, i, i, i, i, i, i],
        'GetRampDataValue':         [i, i, i, i, s, s, d, i],
        'SetIPC_DVM_ID':            [i, i, i, i],
        'GetMEFIValue':             [i, d, d, d, d, i, i, i, i, i],
        'GetMEFIValue_RKA':         [i, d, d, d, d, i],
    })


def _declare(lib, argtypes):
    funcs = {method: lib[method] for method in argtypes}
    for method, types in argtypes.items():
        funcs[method].argtypes = types
        funcs[method].restype = None
    return funcs



"""
Utilities for unicode IO.
"""

import csv
import sys
import time


__all__ = [
    'csv_unicode_reader',
]


if sys.version_info[0] < 3:
    def csv_unicode_reader(lines, encoding='utf-8', **kwargs):
        """Load unicode CSV file."""
        return [[r.decode(encoding) for r in row]
                for row in csv.reader(lines, **kwargs)]

else:
    def csv_unicode_reader(lines, encoding='utf-8', **kwargs):
        """Load unicode CSV file."""
        lines = [l.decode(encoding) for l in lines]
        return csv.reader(lines, **kwargs)


class TimeoutCache(object):

    """
    Cache that cycles every ``timeout`` seconds.

    The invalidation interval is global, not per-entry.

    ``timeout=0`` means no caching,
    ``timeout=-1`` means infinite caching.
    """

    def __init__(self, get, timeout=1.0):
        self._get = get
        self._beg = time.time()
        self.timeout = timeout
        self.values = {}

    def __getitem__(self, name):
        now = time.time()
        beg = self._beg
        timeout = self.timeout
        values = self.values
        if now - beg > timeout > 0:
            self._beg = now - (now - beg) % timeout
            values.clear()
        if name not in values or timeout == 0:
            values[name] = self._get(name)
        return values[name]



"""
Fake implementation of BeamOptikDLL wrapper. Emulates the API of
:class:`~hit_acs.beamoptikdll.BeamOptikDLL`. Primarily used for
offline testing of the basic functionality.
"""

import logging
import functools
from random import gauss, gammavariate as gamma

from pydicti import dicti

from .beamoptikdll import DVMStatus, GetOptions, ExecOptions, GetSDOptions, EFI
from .util import TimeoutCache


__all__ = [
    'BeamOptikStub',
]


def _api_meth(func):
    """Decorator for tracing calls to BeamOptikDLL API methods."""
    @functools.wraps(func)
    def wrapper(self, *args):
        if func.__name__ != 'GetFloatValueSD':
            logging.debug('{}{}'.format(func.__name__, args))
        return func(self, *args)
    return wrapper


class BeamOptikStub(object):

    """
    A fake implementation of the BeamOptikDLL python wrapper for offline
    testing. It does not use the actual DLL but is just a thin simulation
    layer that can partially mimick actual behaviour and provide BPM readings
    based on a provided MAD-X model.
    """

    # TODO: Support read-only/write-only parameters
    # TODO: Prevent writing unknown parameters by default

    def __init__(self, model=None, offsets=None, settings=None, variant='HIT'):
        """Initialize new library instance with no interface instances."""
        if settings is None:
            settings = {}
        self.params = dicti()
        self.sd_values = dicti()
        self.sd_cache = TimeoutCache(
            self._get_jittered_sd,
            timeout=settings.get('shot_interval', 1.0))
        self.model = model
        self.offsets = {} if offsets is None else offsets
        self.settings = settings
        self.jitter = settings.get('jitter', True)
        self.auto_sd = settings.get('auto_sd', True)
        self._variant = variant

    _aberration_magnitude = {
        'ax':  1e-4,    # 0.1 mrad
        'ay':  1e-4,
        'dax': 1e-4,
        'day': 1e-4,
        'kl':  5e-3,    # 0.005
    }

    def _aberrate_strengths(self):
        for k in self.params:
            prefix = k.lower().split('_')[0]
            sigma = self._aberration_magnitude.get(prefix)
            if sigma is not None:
                self.params[k] += gauss(0, sigma)
        self.ExecuteChanges()

    def set_sd_values(self, data):
        self.sd_values = dicti(data)
        self.auto_sd = False

    def set_model(self, model):
        self.model = model
        if model:
            self.set_float_values(model.globals)
            self.update_sd_values()

    def set_float_values(self, data):
        self.params.clear()
        self.params.update({
            'A_POSTSTRIP':  1.007281417537080e+00,
            'Q_POSTSTRIP':  1.000000000000000e+00,
            'Z_POSTSTRIP':  1.000000000000000e+00,
            'E_HEBT':       2.034800000000000e+02,
            # copying HEBT settings for testing:
            'E_SOURCE':     2.034800000000000e+02,
            'E_MEBT':       2.034800000000000e+02,
        })
        self.params.update(data)
        self.ExecuteChanges()

    @_api_meth
    def DisableMessageBoxes(self):
        """Do nothing. There are no message boxes anyway."""
        pass

    @_api_meth
    def GetInterfaceInstance(self):
        """Create a new interface instance."""
        self.vacc = 3
        self.EFIA = (1, 1, 1, 1)
        return 1337

    @_api_meth
    def FreeInterfaceInstance(self):
        """Destroy a previously created interface instance."""
        del self.vacc
        del self.EFIA

    @_api_meth
    def GetDVMStatus(self, status):
        """Get DVM ready status."""
        # The test lib has no advanced status right now.
        return DVMStatus.Ready

    @_api_meth
    def SelectVAcc(self, vaccnum):
        """Set virtual accelerator number."""
        self.vacc = vaccnum

    @_api_meth
    def SelectMEFI(self, vaccnum, energy, focus, intensity, gantry_angle=0):
        """Set MEFI in current VAcc."""
        # The real DLL requires SelectVAcc to be called in advance, so we
        # enforce this constraint here as well:
        assert self.vacc == vaccnum
        self.EFIA = (energy, focus, intensity, gantry_angle)
        return EFI(
            float(energy),
            float(focus),
            float(intensity),
            float(self.params.get('gantry_angle', gantry_angle)))

    @_api_meth
    def GetSelectedVAcc(self):
        """Get currently selected VAcc."""
        return self.vacc

    @_api_meth
    def GetFloatValue(self, name, options=GetOptions.Current):
        """Get a float value from the "database"."""
        return float(self.params.get(name, 0))

    @_api_meth
    def SetFloatValue(self, name, value, options=0):
        """Store a float value to the "database"."""
        self.params[name] = value

    @_api_meth
    def ExecuteChanges(self, options=ExecOptions.CalcDif):
        """Compute new measurements based on current model."""
        if self.model:
            self.model.update_globals(self.params)
            self.update_sd_values()

    @_api_meth
    def SetNewValueCallback(self, callback):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def GetFloatValueSD(self, name, options=GetSDOptions.Current):
        """Get beam diagnostic value."""
        try:
            storage = self.sd_cache if self.jitter else self.sd_values
            return storage[name] * 1000
        except KeyError:
            return -9999.0

    def _get_jittered_sd(self, name):
        value = self.sd_values[name]
        prefix = name.lower().split('_')[0]
        if value != -9999:
            mean, stddev = value, 1e-4      # 0.1 mm
            if prefix in ('posx', 'posy'):
                return gauss(mean, stddev)
            if prefix in ('widthx', 'widthy'):
                return gamma(mean**2/stddev**2, stddev**2/mean)
        return value

    def update_sd_values(self):
        """Compute new measurements based on current model."""
        model = self.model
        if not model or not self.auto_sd:
            return
        model.twiss()
        for elem in self._monitors():
            dx, dy = self.offsets.get(elem.name, (0, 0))
            twiss = model.get_elem_twiss(elem.name)
            values = {
                'widthx': twiss.envx,
                'widthy': twiss.envy,
                'posx': -twiss.x - dx,
                'posy': twiss.y - dy,
            }
            self.sd_values.update({
                key + '_' + elem.name: val
                for key, val in values.items()
            })

    @_api_meth
    def GetLastFloatValueSD(self, name, vaccnum,
                            energy, focus, intensity, gantry_angle=0,
                            options=GetSDOptions.Current):
        """Get beam diagnostic value."""
        # behave exactly like GetFloatValueSD and ignore further parameters
        # for now
        return self.GetFloatValueSD(name, options)

    @_api_meth
    def StartRampDataGeneration(self, vaccnum, energy, focus, intensity):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def GetRampDataValue(self, order_num, event_num, delay,
                         parameter_name, device_name):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def SetIPC_DVM_ID(self, name):
        """Not implemented."""
        raise NotImplementedError

    @_api_meth
    def GetMEFIValue(self):
        """Get current MEFI combination."""
        channels = EFI(*self.EFIA)
        values = EFI(
            float(channels.energy),
            float(channels.focus),
            float(channels.intensity),
            float(self.params.get('gantry_angle', channels.gantry_angle)))
        return (values, channels)

    def _monitors(self):
        return [elem for elem in self.model.elements
                if elem.base_name.endswith('monitor')]
